####################################################################
#
#   MarathUp - MacBinary file output
#
####################################################################

module Marathon

	class MacBinaryFile
	
		def MacBinaryFile.create(path, type, creator)
			f = File.open(path, "wb+")
			mb = MacBinaryFile.new(f, true)
			mb.write_header(File.basename(path), type, creator)
			return mb
		end
		
		def MacBinaryFile.open_data_fork(path)
			f = File.open(path, "rb")
			f.seek(102)
			signature = f.read(4)
			f.seek(0)
			if signature == "mBIN"
				printf("Opening MacBinary file %s for reading\n", path) ###
				mb = MacBinaryFile.new(f, false)
				mb.seek_data_fork()
				return mb
			else
				printf("Opening plain file %s for reading\n", path) ###
				return f
			end
		end
		
		def initialize(f, writing)
			@f = f
			@writing = writing
		end
		
		def tell()
			return @f.tell() - @fork_start
		end
		
		def seek(whither, whence = 0)
			case whence
				when 0
					@f.seek(@fork_start + whither)
				when 1
					@f.seek(whither, 1)
				when 2
					if not @writing
						@f.seek(@fork_end + whither)
					else
						@f.seek(whither, 2)
					end
			end
		end
	
		def close()
			if @writing
				@f.flush()
				@f.seek(0)
				header = @f.read(124)
				crc = macbinary_crc(header)
				#printf("Writing crc 0x%04X\n", crc)
				@f.seek(124)
				@f.write([crc].pack("n"))
			end
			@f.close()
		end
		
		#------------------------------------------------------------------
		#
		#   Reading
		#		
		#------------------------------------------------------------------

		def seek_data_fork()
			@f.seek(83)
			length = @f.read(4).unpack("N")[0]
			@fork_start = 128
			@fork_end = @fork_start + length
			@f.seek(@fork_start)
			printf("Data fork start 0x%x end 0x%x\n", @fork_start, @fork_end) ###
		end
		
		def read(length = nil)
			here = @f.tell()
			if not length or here + length > @fork_end
				length = @fork_end - here
			end
			return @f.read(length)
		end
		
		#------------------------------------------------------------------
		#
		#   Writing
		#		
		#------------------------------------------------------------------

		def write_header(name, type, creator)
			name = name[0, 63]
			data = [
				0,           #  000  Byte  Old version number, must be kept at zero for compatibility
				name.length, #  001  Byte  Length of filename (must be in the range 1-31)
				name,        #  002  Byte[63] Filename (only "length" bytes are significant).
				type,        #  065  Long  File type (normally expressed as four characters)
				creator,     #  069  Long  File creator (normally expressed as four characters)
				0,           #  073  Byte  Original Finder flags
				#                            Bit 7 - isAlias.
				#                            Bit 6 - isInvisible.
				#                            Bit 5 - hasBundle.
				#                            Bit 4 - nameLocked.
				#                            Bit 3 - isStationery.
				#                            Bit 2 - hasCustomIcon.
				#                            Bit 1 - reserved.
				#                            Bit 0 - hasBeenInited.
				0,           #  074  Byte  Zero fill, must be zero for compatibility
				0,           #  075  Word  File's vertical position within its window.
				0,           #  077  Word  File's horizontal position within its window.
				0,           #  079  Word  File's window or folder ID.
				0,           #  081  Byte  "Protected" flag (in low order bit).
				0,           #  082  Byte  Zero fill, must be zero for compatibility
				0,           #  083  Long  Data Fork length (bytes, zero if no Data Fork).
				0,           #  087  Long  Resource Fork length (bytes, zero if no R.F.).
				0,           #  091  Long  File's creation date
				0,           #  095  Long  File's "last modified" date.
				0,           #  099  Word  Length of Get Info comment to be sent after the resource fork (if implemented, see below).
				0,           #  101  Byte  Finder Flags, bits 0-7. (Bits 8-15 are already in byte 73)
				#                            Bit 7 - hasNoInits
				#                            Bit 6 - isShared
				#                            Bit 5 - requiresSwitchLaunch
				#                            Bit 4 - ColorReserved
				#                            Bits 1-3 - color
				#                            Bit 0 - isOnDesk
				'mBIN',      # *102  Long  Signature for indentification purposes ('mBIN')
				0,           # *106  Byte  Script of file name (from the fdScript field of an fxInfo record)
				0,           # *107  Byte  Extended Finder flags (from the fdXFlags field of an fxInfo record)
				"",          #  108-115    Unused (must be zeroed by creators, must be ignored by readers)
				0,           #  116  Long  Length of total files when packed files are unpacked. As of the writing of this document, this field has never been used.
				0,           #  120  Word  Length of a secondary header. If this is non-zero, skip this many bytes (rounded up to the next multiple of 128). This is for future expansion only, when sending files with MacBinary, this word should be zero.
				0,           # *122  Byte  Version number of MacBinary III that the uploading program is written for (the version is 130 for MacBinary III)
				0,           #  123  Byte  Minimum MacBinary version needed to read this file (set this value at 129 for backwards compatibility with MacBinary II)
				0,           #  124  Word  CRC of previous 124 bytes
				"",          #  126        Padding to 128 bytes
			]
			bytes = data.pack("cca63a4a4ccnnnccNNNNnca4cca8Nnccna2")
			if bytes.length != 128
				raise StandardError.new(format("Header bytes wrong length (%s)", bytes.length))
			end
			@f.write(bytes)
		end
		
		def open_data_fork()
			@fork_start = @f.tell()
		end
		
		def close_data_fork()
			self.close_fork(83)
		end
		
		def open_resource_fork()
			@fork_start = @f.tell()
		end
		
		def close_resource_fork()
			self.close_fork(87)
		end
		
		def close_fork(len_offset)
			here = self.pad_to_block()
			len = here - @fork_start
			len_bytes = [len].pack("N")
			#printf("Fork started at 0x%x\n", @fork_start) ###
			#printf("Fork ends at    0x%x\n", here) ###
			#printf("Fork length     0x%x\n", len) ###
			#printf("Writing         %s\n", len_bytes.inspect) ###
			@f.flush()
			@f.seek(len_offset)
			@f.write(len_bytes)
			@f.flush()
			@f.seek(0, 2)
		end
		
		def pad_to_block()
			pos = @f.tell()
			pad = (0x80 - (pos & 0x7f)) & 0x7f
			#printf("Writing %s pad bytes\n", pad) ###
			@f.write("\0" * pad)
			return pos
		end
		
		def write(data)
			#printf("Writing 0x%x bytes\n", data.length) ###
			@f.write(data)
			#printf("Position now 0x%x\n", @f.tell()) ###
		end
		
		def flush()
			@f.flush()
		end
		
		def macbinary_crc(header)
			crc = 0
			for i in 0...124
				data = header[i] << 8
				for j in 0...8
					if ((data ^ crc) & 0x8000) != 0
						crc = ((crc << 1) ^ 0x1021) & 0xffff
					else
						crc = (crc << 1) & 0xffff
					end
					data <<= 1
				end
			end
			return crc
		end
		
	end # class MacBinary

end # module Marathon
