##################################################################################
#
#   MarathUp - Map file output
#
##################################################################################

module Marathon

	#------------------------------------------------------------------
	#
	#   Map file output utilities
	#
	#------------------------------------------------------------------

	def Marathon.find_resource_file(dir, name)
		return find_related_file(dir, name, ".rsrc")
	end
	
	def Marathon.find_physics_file(dir, name)
		return find_related_file(dir, name, ".phys")
	end
	
	def Marathon.find_related_file(dir, name, extension)
		candidates = [name, "Map"]
		for name in candidates
			path = File.join(dir, name + extension)
			printf("Looking for %s\n", path) ###
			begin
				File.stat(path)
				return path
			rescue SystemCallError
				#printf("Rescued %s\n", $!.class.name) ###
			end
		end
		return nil
	end

	def Marathon.copy_resource_fork(f, rsrc_path)
		g = File.open(rsrc_path + "/rsrc", "rb")
		rsrc_data = g.read()
		g.close()
		f.open_resource_fork()
		f.write(rsrc_data)
		f.close_resource_fork()
	end
			
	def Marathon.write_map_header(f, name, num_levels, with_app_specific_data)
		if with_app_specific_data
			asds = 0x4a
		else
			asds = 0
		end
		data = [
			2,           # short version
			1,           # short data_version
			name,        # char file_name[64]
			0,           # unsigned long checksum
			0,           # long directory_offset
			num_levels,  # short wad_count
			asds,        # short application_specific_directory_data_size
			16,          # short entry_header_size
			10,          # short directory_entry_base_size
			0,           # unsigned long parent_checksum
			""]          # short unused[20]
		write_struct(f, "nna64NNnnnnNa40", data, "header", 128)
	end

	def Marathon.write_map_directory_entry(f, offset, length, index, app_specific_data)
		base_data = [
			offset,  # long offset_to_start
			length,  # long length
			index]   # short index
#		app_specific_data = [
#			0,       # short mission_flags
#			0,       # short environment_flags
#			0x01,    # long entry_point_flags
#			name]    # char level_name[66]
#			# Official maps have 0xd8 in last byte of level_name???
		write_struct(f, "NNn", base_data, "base directory entry", 10)
		if app_specific_data
			write_struct(f, "nnNa66", app_specific_data, "app specific directory data", 0x4a)
		end
	end

	def Marathon.fix_map_directory_offset(f, offset)
		data = [offset]
		f.flush()
		f.seek(72, 0)
		write_struct(f, "N", data, "directory offset", 4)
		f.seek(0, 2)
	end
	
	def Marathon.read_map_chunk_header(f)
		bytes = f.read(16)
		if not bytes or bytes.length < 16
			raise Marathon::ExportError.new("Unexpected EOF reading chunk header")
		end
		return bytes.unpack("a4N3")
	end
	
	def Marathon.write_map_chunk_header(f, level_start, tag, length, last = false)
		start_offset = f.tell() - level_start
		if not last
			next_offset = start_offset + 16 + length
		else
			next_offset = 0
		end
		header = [
			tag,         # WadDataType tag
			next_offset, # long next_offset
			length,      # long length
			0]           # long offset
		write_struct(f, "a4NNN", header, "chunk header", 16)
	end
	
	def Marathon.write_struct(f, format, data, name, length)
		#printf("Writing struct: %s %s\n", name, length) ###
		#printf("Format = %s\n", format.inspect) ###
		#printf("Data = %s\n", data.inspect) ###
		begin
			bytes = data.pack(format)
		rescue
			printf("Error packing data: %s\n", data.inspect)
			raise
		end
		check_length(bytes, name, length)
		f.write(bytes)
	end
	
	def Marathon.check_length(bytes, name, length)
		if bytes.length != length
			raise "Wrong length packed data for " + name
		end
	end

end # module marathon

class MarathonExporter

	#------------------------------------------------------------------
	#
	#   Map File Output
	#
	#   Write the generated Marathon objects to a map file.
	#
	#------------------------------------------------------------------

	def write_map_file(model_path)
		dir = File.dirname(model_path)
		name = File.basename(model_path, ".skp")
		map_path = File.join(dir, name + ".sceA")
		rsrc_path = Marathon.find_resource_file(dir, name)
		phys_path = Marathon.find_physics_file(dir, name)
		if rsrc_path
			self.write_macbinary_map_file(name, map_path, phys_path, rsrc_path)
		else
			self.write_plain_map_file(name, map_path, phys_path)
		end
		UI.messagebox(format("Written Marathon map to '%s'.", map_path))
	end

	def write_plain_map_file(name, map_path, phys_path)
		#printf("Writing plain map file to %s\n", map_path) ###
		f = File.open(map_path, "wb")
		self.write_map_data(f, name, phys_path)
		f.close()
	end
	
	def write_macbinary_map_file(name, map_path, phys_path, rsrc_path)
		#printf("Writing MacBinary map file to %s\n", map_path) ###
		#printf("Merging resources from %s\n", rsrc_path) ###
		f = Marathon::MacBinaryFile.create(map_path, "sce2", "26.\xb0")
		f.open_data_fork()
		self.write_map_data(f, name, phys_path)
		f.close_data_fork()
		Marathon.copy_resource_fork(f, rsrc_path)
		f.close()
	end
	
	def write_map_data(f, name, phys_path)
		self.write_header(f, name)
		self.write_map_info(f, name)
		self.write_points(f)
		self.write_lines(f)
		self.write_sides(f)
		self.write_polygons(f)
		self.write_platforms(f)
		self.write_objects(f)
		self.write_lights(f)
		self.write_media(f)
		self.write_ambient_sounds(f)
		self.write_random_sounds(f)
		self.write_notes(f)
		self.write_terminals(f)
		self.write_placement_data(f, phys_path == nil) # Must be called last - terminates list of chunks
		if phys_path
			self.copy_physics(f, phys_path)
		end
		dir_offset = self.write_directory_entry(f, name)
		self.fix_directory_offset(f, dir_offset)
	end
	
	def write_header(f, name)
		Marathon.write_map_header(f, name, 1, false)
	end
	
	def write_directory_entry(f, name)
		directory_offset = f.tell()
		Marathon.write_map_directory_entry(f, 128, directory_offset - 128, 0, nil)
		return directory_offset
	end
	
	def fix_directory_offset(f, offset)
		Marathon.fix_map_directory_offset(f, offset)
	end
	
	def write_map_info(f, default_name)
		length = 88
		write_chunk_header(f, "Minf", length)
		attrs = Marathon.get_level_attributes()
		env_code = (attrs["environment_code"] or 0)
		physics = (attrs["physics_model"] or 0)
		song = (attrs["song_index"] or 0)
		mission = (attrs["mission_flags"] or 0)
		env_flags = (attrs["environment_flags"] or 0)
		level_name = (attrs["level_name"] or "")
		if level_name == ""
			level_name = default_name
		end
		entry = (attrs["entry_point_flags"] or 0)
		if entry == 0
			entry = 0x1 # single player
		end
		data = [
			env_code,    # TODO: short environment_code;
			physics,     # TODO: short physics_model;
			song,        # TODO: short song_index;
			mission,     # TODO: short mission_flags;
			env_flags,   # TODO: short environment_flags;
			"",          # short unused[4];
			level_name,  # char level_name[LEVEL_NAME_LENGTH];
			entry]       # TODO: long entry_point_flags;
		write_struct(f, "n5a8a66N", data, "map info", length)
	end
	
	def prefixed_attr(dict, prefix, name)
		key = format("%s.%s", prefix, name)
		return dict[key]
	end
	
	def fixed(x)
		return (x * 0x10000).round()
	end
	
	def light_phase_bytes(attrs, prefix, default_period = 0)
		f = (prefixed_attr(attrs, prefix, "function") or 0)
		p = (prefixed_attr(attrs, prefix, "period") or 0)
		dp = (prefixed_attr(attrs, prefix, "delta_period") or 0)
		i = (prefixed_attr(attrs, prefix, "intensity") or 0.0)
		di = (prefixed_attr(attrs, prefix, "delta_intensity") or 0.0)
		if p == 0
			p = default_period
		end
		data = [
			f,         # short function;
			p,         # short period;
			dp,        # short delta_period;
			fixed(i),  # fixed intensity;
			fixed(di), # fixed delta_intensity;
		]
		if @@DebugLights
			printf("Light phase %s data = %s\n", prefix, data.inspect)
		end
		return data.pack("n3N2")
	end
	
	def write_lights(f)
		n = @lights.length
		length = 100
		write_chunk_header(f, "LITE", n * length)
		for light in @lights
			attrs = light.attrs
			type = (attrs["type"] or 0)
			flags = (attrs["flags"] or 0)
			phase = (attrs["phase"] or 0)
			tag = self.tag_named(light.tag_name)
			pa = light_phase_bytes(attrs, "primary_active", 30)
			sa = light_phase_bytes(attrs, "secondary_active")
			ba = light_phase_bytes(attrs, "becoming_active")
			pi = light_phase_bytes(attrs, "primary_inactive", 30)
			si = light_phase_bytes(attrs, "secondary_inactive")
			bi = light_phase_bytes(attrs, "becoming_inactive")
			data = [
				type,   # short type;
				flags,  # word flags;
				phase,  # short phase;
				pa,     # struct lighting_function_specification primary_active, 
				sa,     # struct lighting_function_specification secondary_active, 
				ba,     # struct lighting_function_specification becoming_active;
				pi,     # struct lighting_function_specification primary_inactive,
				si,     # struct lighting_function_specification secondary_inactive,
				bi,     # struct lighting_function_specification becoming_inactive;
				tag,    # short tag;
				""]     # short unused[4];
			if @@DebugLights
				printf("Writing light %s\n", attrs.inspect)
				printf("Light data = %s\n", data.inspect)
			end
			write_struct(f, "n3a14a14a14a14a14a14na8", data, "LITE entry", length)
		end
	end
	
	def write_media(f)
		n = @media.length
		length = 32
		write_chunk_header(f, "medi", n * length)
		for media, trans in @media
			begin
				inst = media.inst
				ctx = media.ctx
				trans = ctx.trans
				attrs = Marathon.get_media_attributes(inst)
				box = inst.bounds
				low_tide = marathon_coord((trans * box.min).z)
				high_tide = marathon_coord((trans * box.max).z)
				type = (attrs["type"] or 0)
				flags = (attrs["flags"] or 0)
				light_index = self.number_of_light_named(media.tide_name)
				p, flow_dir = position_and_facing(inst, ctx)
				current = (attrs["current_magnitude"] or 0.0)
				flow_speed = marathon_speed(current)
			rescue Marathon::ExportError
				error(media, trans, $!.message)
			end
			data = [
				type,        # short type;
				flags,       # word flags;
				light_index, # short light_index;
				flow_dir,    # angle current_direction;
				flow_speed,  # world_distance current_magnitude;	
				low_tide,    # world_distance low;
				high_tide,   # world_distance high;	
				0, 0,        # world_point2d origin;
				0,           # world_distance height;
				0,           # fixed minimum_light_intensity;
				-1,          # shape_descriptor texture;
				0,           # short transfer_mode;	
				""]          # short unused[2];
			write_struct(f, "n10Nn2a4", data, "medi entry", length)
		end
	end
	
	def marathon_volume(x)
		return (x * 0x100).round()
	end

	def write_ambient_sounds(f)
		n = @ambient_sounds.length
		length = 16
		write_chunk_header(f, "ambi", n * length)
		for ambi, trans in @ambient_sounds
			attrs = Marathon.get_ambient_sound_attributes(ambi)
			index = (attrs["sound_index"] or 0)
			volume = marathon_volume((attrs["volume"] or 0))
			data = [
				0,      # word flags;
				index,  # short sound_index;
				volume, # short volume;
				""]     # short unused[5];
			write_struct(f, "n3a10", data, "ambi entry", length)
		end
	end
	
	def write_random_sounds(f)
		n = @random_sounds.length
		length = 32
		write_chunk_header(f, "bonk", n * length)
		for bonk, trans in @random_sounds
			attrs = Marathon.get_random_sound_attributes(bonk)
			flags = (attrs["flags"] or 0)
			index = (attrs["sound_index"] or 0)
			volume = marathon_volume((attrs["volume"] or 0.5))
			d_volume = marathon_volume((attrs["delta_volume"] or 0.0))
			period = (attrs["period"] or 60)
			d_period = (attrs["delta_period"] or 0)
			pitch = (attrs["pitch"] or 1.0)
			d_pitch = (attrs["delta_pitch"] or 0.0)
			p, facing = position_and_facing(bonk, trans)
			d_facing = ((attrs["delta_direction"] or 0.0) * 0x200 / 360).round()
			data = [
				flags,           # word flags;
				index,           # short sound_index;
				volume,          # short volume;
				d_volume,        # short delta_volume;
				period,          # short period;
				d_period,        # short delta_period;
				facing,          # angle direction;
				d_facing,        # angle delta_direction;
				fixed(pitch),    # fixed pitch;
				fixed(d_pitch),  # fixed delta_pitch;
				0,               # short phase; // only used at run-time; initialize to NONE
				""]              # short unused[3];
			write_struct(f, "n8N2na6", data, "bonk entry", length)
		end
	end
	
	def write_notes(f)
		n = @notes.length
		length = 72
		write_chunk_header(f, "NOTE", n * length)
		for note, trans in @notes
			p, a = position_and_facing(note, trans)
			xy = marathon_coords([p.x, p.y])
			z = marathon_coord(p.z)
			poly = self.find_marathon_poly(xy, z)
			if poly
				attrs = Marathon.get_note_attributes(note)
				text = (attrs["text"]	or "")
				x, y = xy
				data = [
					0,           # short type; # unused, must be 0
					x, -y,       # world_point2d location;
					poly.number, # short polygon_index;
					text]        # char text[MAXIMUM_ANNOTATION_TEXT_LENGTH];
				write_struct(f, "n4a64", data, "note entry", length)
			else
				error(note, trans, "Unable to find containing polygon for note.")
			end
		end
	end	

	def write_points(f)
		data = []
		for x, y in @marathon_points
			data.push([x, -y].pack("nn"))
		end
		write_chunk(f, "PNTS", data)
	end
	
	def write_lines(f)
		length = @marathon_lines.length * 32
		write_chunk_header(f, "LINS", length)
		for line in @marathon_lines
			p = line.point_numbers
			data = [
				p[0], p[1],                           # short endpoint_indexes[2]
				line.flags,                           # word flags
				line.length(@marathon_points),        # world_distance length
				line.highest_adjacent_floor,          # world_distance highest_adjacent_floor
				line.lowest_adjacent_ceiling,         # world_distance lowest_adjacent_ceiling
				number(line.sides[CLOCKWISE]),        # short clockwise_polygon_side_index
				number(line.sides[ANTICLOCKWISE]),    # short counterclockwise_polygon_side_index
				number(line.polygons[CLOCKWISE]),     # short clockwise_polygon_owner
				number(line.polygons[ANTICLOCKWISE]), # short counterclockwise_polygon_owner
				""] #	short unused[6];
			write_struct(f, "nnnnnnnnnna12", data, "LINS element", 32)
		end
	end
	
	def control_panel_permutation(control_panel_type, name)
		if control_panel_type >= 0
			types = Marathon.get_control_panel_types()
			perm_type = types[control_panel_type][1]
			#printf("Getting control panel permutation for %s\n", name.inspect) ###
			#printf("Permutation type = %s\n", perm_type.inspect) ###
			case perm_type
				when 'platform'
					#printf("Getting platform named %s\n", name.inspect) ###
					return number(self.poly_named(name))
				when 'light'
					return self.number_of_light_named(name)
				when 'tag'
					return self.tag_named(name)
				when 'terminal'
					#printf("Getting terminal number for terminal file name %s\n",
					#	name.inspect) ###
					n = number(self.terminal_named(name))
					#printf("Terminal number = %s\n", n.inspect) ###
					return n
			end
		end
		return 0
	end
	
	def write_sides(f)
		length = @marathon_sides.length * 64
		write_chunk_header(f, "SIDS", length)
		for side in @marathon_sides
			begin
				attrs = side.attributes()
				control_panel_type = (attrs["control_panel_type"] or -1)
				perm_name = side.permutation_name
				#if control_panel_type >= 0
				#	printf("Control panel type %s perm_name %s\n",
				#		control_panel_type.inspect, perm_name.inspect) ###
				#end
				permutation = self.control_panel_permutation(control_panel_type, perm_name)
				primary_texture = side.primary_texture
				secondary_texture = side.secondary_texture
				transparent_texture = side.transparent_texture
				if @@DebugLights
					printf("Writing lights of side %s\n", side.number)
				end
				primary_light = number_of_light_named(side.primary_light)
				secondary_light = number_of_light_named(side.secondary_light)
				transparent_light = number_of_light_named(side.transparent_light)
				if control_panel_type < 0
					control_panel_type = 0
				end
				data = [
					side.type,                           # short type
					side.flags,                          # word flags
					tex_bytes(primary_texture),          # struct side_texture_definition primary_texture
					tex_bytes(secondary_texture),        # struct side_texture_definition secondary_texture
					tex_bytes(transparent_texture),      # struct side_texture_definition transparent_texture
					0, 0, 0, 0,                          # struct side_exclusion_zone exclusion_zone
					control_panel_type,                  # short control_panel_type
					permutation,                         # short control_panel_permutation
					primary_texture.transfer_mode,       # short primary_transfer_mode
					secondary_texture.transfer_mode,     # short secondary_transfer_mode
					transparent_texture.transfer_mode,   # short transparent_transfer_mode
					number(side.polygon),                # short polygon_index
					number(side.line),                   # short line_index
					primary_light,                       # short primary_lightsource_index
					secondary_light,                     # short secondary_lightsource_index
					transparent_light,                   # short transparent_lightsource_index
					0,                                   # TODO: fixed ambient_delta
					""]                                  # short unused[1]
			rescue Marathon::ExportError
				wall = side.line.representative_wall
				error(wall.face, wall.ctx, $!.message)
			end
			write_struct(f, "nna6a6a6N4nnnnnnnnnnNa2", data, "SIDS element", 64)
		end
	end
	
	def tex_bytes(obj)
		if obj
			data = [obj.x0, obj.y0, shape_bytes(obj.texture)]
			return data.pack("nna2")
		else
			return MarathonSideTextureDefinition::NullBytes
		end
	end
	
	def shape_bytes(obj)
		if obj
			data = [obj.set, obj.frame]
			return data.pack("CC")
		else
			return MarathonShapeDescriptor::NullBytes
		end
	end
	
	def poly_permutation_index(type, name)
		case type
			when 4 # _polygon_is_base
				return name.to_i()
			when 6, 8 # _polygon_is_light_on_trigger, _polygon_is_light_off_trigger
				return number(self.light_named(name))
			when 7, 9, 10 # _polygon_is_platform_on_trigger, _polygon_is_platform_off_trigger, _polygon_is_teleporter
				return number(self.poly_named(name))
			else
				return 0
		end
	end

	def write_polygons(f)
		length = @marathon_polys.length * 128
		write_chunk_header(f, "POLY", length)
		obj_index = 0
		for poly in @marathon_polys
			begin
				attrs = Marathon.get_polygon_attributes(poly.floor)
				type = (attrs["type"] or 0)
				if type == 5 # _polygon_is_platform
					permutation = poly.platform.number
				else
					permutation = self.poly_permutation_index(type, poly.permutation_name)
				end
				points = Array.new(poly.point_numbers)
				for i in points.length...8
					points.push(-1)
				end
				if @@DebugLights
					printf("Writing lights of poly %s\n", poly.number)
				end
				floor_light = self.number_of_light_named(poly.floor_light)
				ceiling_light = self.number_of_light_named(poly.ceiling_light)
				floor_tex = poly.floor_texture
				ceiling_tex = poly.ceiling_texture
				floor_x0, floor_y0 = poly.floor_origin
				ceiling_x0, ceiling_y0 = poly.ceiling_origin
				if floor_tex and floor_tex.is_landscape
					floor_tm = 9 # landscape
				else
					floor_tm = poly.floor_transfer_mode
				end
				if ceiling_tex and ceiling_tex.is_landscape
					ceiling_tm = 9 # landscape
				else
					ceiling_tm = poly.ceiling_transfer_mode
				end
				media_index = self.number_of_media_named(poly.media_name)
				media_light_index = self.number_of_light_named(poly.media_light)
				ambient_sound_index = self.number_of_ambient_sound_named(poly.ambient_sound)
				random_sound_index = self.number_of_random_sound_named(poly.random_sound)
				flags = 0
				data = [
					type,                              # short type;
					flags,                             # TODO: word flags;
					permutation,                       # short permutation;
					poly.num_points,                   # short vertex_count;
					points.pack("n8"),                 # short endpoint_indexes[MAXIMUM_VERTICES_PER_POLYGON];
					numbers(poly.lines).pack("n8"),    # short line_indexes[MAXIMUM_VERTICES_PER_POLYGON];
					shape_bytes(floor_tex),            # shape_descriptor floor_texture;
					shape_bytes(ceiling_tex),          # shape_descriptor ceiling_texture;
					poly.floor_height,                 # world_distance floor_height;
					poly.ceiling_height,               # world_distance ceiling_height;
					floor_light,                       # short floor_lightsource_index;
					ceiling_light,                     # short ceiling_lightsource_index;
					[poly.area].pack("N"),             # long area;
					obj_index,                         # short first_object;
					0,                                 # short first_exclusion_zone_index;
					0,                                 # short line_exclusion_zone_count;
					0,                                 # short point_exclusion_zone_count;
					floor_tm,                          # short floor_transfer_mode;
					ceiling_tm,                        # short ceiling_transfer_mode;
					numbers(poly.adjacent_polys).pack("n8"), # short adjacent_polygon_indexes[MAXIMUM_VERTICES_PER_POLYGON];
					0,                                 # short first_neighbor_index;
					0,                                 # short neighbor_count;
					poly.center.pack("n2"),            # world_point2d center;
					numbers(poly.sides).pack("n8"),    # short side_indexes[MAXIMUM_VERTICES_PER_POLYGON];
					floor_x0, -floor_y0,               # world_point2d floor_origin;
					ceiling_x0, -ceiling_y0,           # world_point2d ceiling_origin;
					media_index,                       # short media_index;
					media_light_index,                 # short media_lightsource_index;
					0,                                 # short sound_source_indexes;
					ambient_sound_index,               # short ambient_sound_image_index;
					random_sound_index,                # short random_sound_image_index;
					""]                                # short unused[1];
			rescue Marathon::ExportError
				error(poly.ceiling, poly.ceiling_trans, $!.message)
			end
			#printf("POLY data = %s\n", data.inspect) ###
			#for i in 0...data.length
			#	if data[i] == nil
			#		printf("data[%s] == nil\n", i)
			#	end
			#end
			write_struct(f, "n4a16a16a2a2n4a4n6a16n2a4a16n4n5a2", data, "POLY element", 128)
			obj_index += poly.objects.length
		end
	end
	
	def write_platforms(f)
		length = @marathon_platforms.length * 32
		write_chunk_header(f, "plat", length)
		for plat in @marathon_platforms
			begin
				attrs = plat.attributes
				speed = marathon_speed((attrs["speed"] or 0))
				tag = self.tag_named(plat.tag_name)
				data = [
					(attrs["type"] or 0),         # short type;
					speed,                        # short speed;
					(attrs["delay"] or 0),        # short delay;
					-1,                           # world_distance maximum_height;
					-1,                           # world_distance minimum_height;
					(attrs["static_flags"] or 0), # unsigned long static_flags;
					plat.polygon.number,          # short polygon_index;
					tag,                          # short tag;
					""]                           # short unused[7];
			rescue Marathon::ExportError
				poly = plat.polygon
				error(poly.ceiling, poly.ceiling_trans, $!.message)
			end
			write_struct(f, "n5Nn2a14", data, "plat element", 32)
		end
	end
	
	def write_objects(f)
		length = @marathon_objects.length * 16
		write_chunk_header(f, "OBJS", length)
		for poly in @marathon_polys
			if @@DebugObjects
				printf("Writing objects for polygon %s: %s\n", poly.number, poly.objects.inspect)
			end
			for obj in poly.objects
				if @@DebugObjects
					printf("Writing object %s\n", obj.inspect)
				end
				begin
					type = obj.type
					flags = obj.flags
					poly = obj.polygon
					(x, y), z = obj.location
					if flags & 0x02 != 0 # _map_object_hanging_from_ceiling
						dz = z - poly.ceiling_height
					else
						dz = z - poly.floor_height
					end
					loc = [x, -y, dz].pack("n3")
					if type == 5 # sound
						attrs = Marathon.get_object_attributes(obj.inst)
						volume = (attrs["volume"] or "")
						if (attrs["volume_mode"] or 0) == 0
							facing = marathon_volume(volume.to_f())
							if facing < 0
								facing = 1
							end
						else
							facing = -number_of_light_named(volume)
						end
					else
						facing = obj.facing
					end
				rescue Marathon::ExportError
					error(obj.inst, obj.trans, $!.message)
				end
				data = [
					type,            # short type;
					obj.index,       # short index;
					facing,          # short facing;
					poly.number,     # short polygon_index;
					loc,             # world_point3d location; // .z is a delta
					flags]           # word flags;
				write_struct(f, "n4a6n1", data, "OBJS element", 16)
			end
		end
	end
	
	def terminal_permutation_index(type, value)
		case type
			when 6 # _interlevel_teleport
				if value == "end"
					return 256
				else
					return value.to_i() - 1
				end
			when 7 # _intralevel_teleport_group
				return number(self.poly_named(value))
			when 8 # _checkpoint_group
				return self.goal_named(value)
			when 14 # _camera_group
				return number(self.object_named(value))
			else
				return value.to_i()
		end
	end
	
	def calculate_term_chunk_length()
		length = 0
		for term in @marathon_terminals
			length += term.total_length
		end
		return length
	end
	
	def write_terminals(f)
		length = self.calculate_term_chunk_length()
		write_chunk_header(f, "term", length)
		for term in @marathon_terminals
			if Marathon::DebugTerminal
				printf("Writing terminal %s\n", term)
			end
			begin
				data = [
					term.total_length, # short total_length;
					term.flags, # short flags;
					term.lines_per_page, # short lines_per_page;
					term.group_count, # short grouping_count;
					term.font_changes_count, # short font_changes_count;
				]
				write_struct(f, "n5", data, "term header", 10)
				for group in term.groups
					permutation = terminal_permutation_index(group.type, group.permutation)
					length = group.length
					mlc = group.maximum_line_count
					data = [
						group.flags, # short flags; /* varies.. */
						group.type, # short type; /* _information_text, _checkpoint_text, _briefing_text, _movie, _sound_bite, _soundtrack */
						permutation, # short permutation; /* checkpoint id for chkpt, level id for _briefing, movie id for movie, sound id for sound, soundtrack id for soundtrack */
						group.start_index, # short start_index;
						length, # short length;
						mlc, # short maximum_line_count;
					]
					write_struct(f, "n6", data, "term group", 12)
				end
				for font in term.font_changes
					data = [
						font.index, # short index;
						font.face, # short face;
						font.color, # short color;
					]
					write_struct(f, "n3", data, "term font", 6)
				end
				rescue Marathon::ExportError
					wall = term.representative_line.representative_wall
					error(wall.face, wall.trans, $!.message)
				end
				f.write(term.text)
			end
		end
	
	def write_placement_data(f, last)
		length = 128 * 12
		write_chunk_header(f, "plac", length, last)
		item_attrs = Marathon.get_item_placement_attributes()
		monster_attrs = Marathon.get_monster_placement_attributes()
		write_placement_records(f, @item_counts, item_attrs)
		write_placement_records(f, @monster_counts, monster_attrs)
	end
	
	def write_placement_records(f, counts, attrs)
		length = 12
		lists = {
			"ini" => [nil] * 64,
			"min" => [0] * 64,
			"max" => [0] * 64,
			"tot" => [0] * 64,
			"app" => [0] * 64,
			"inf" => [false] * 64,
			"rnd" => [false] * 64,
		}
		for key, value in attrs
			if value
				row, col = key.split(".")
				#printf("Recombifying %s, %s => %s\n", row.inspect, col.inspect, value.inspect) ###
				lists[col][row.to_i()] = value
			end
		end
		for i in 0...64
			flags = 0
			initial = (lists["ini"][i] or counts[i])
			minimum = lists["min"][i]
			maximum = lists["max"][i]
			total = lists["tot"][i]
			chance = 0xffff * lists["app"][i] / 100
			if lists["inf"][i]
				total = 0xffff
			end
			if lists["rnd"][i]
				flags |= 0x1
			end
			data = [
				flags,    # word flags;
				initial,  # short initial_count;
				minimum,  # short minimum_count;
				maximum,  # short maximum_count;
				total,    # short random_count;
				chance]   # word random_chance;
			write_struct(f, "n6", data, "placement record", length)
		end
	end
	
	PhysicsTags = ['MNpx', 'FXpx', 'PRpx', 'PXpx', 'WPpx']

	def copy_physics(f, phys_path)
		begin
			g = File.open(phys_path, "rb")
			g.seek(128)
			while true
				header = read_chunk_header(g)
				tag, next_offset, length, offset = header
				if not PhysicsTags.index(tag)
					raise Marathon::ExportError.new(format(
						"Invalid physics chunk type '%s'", tag))
				end
				#printf("Copying physics chunk %s length 0x%x next_offset 0x%x\n",
				#	tag.inspect, length, next_offset) ###
				last = (next_offset == 0)
				data = g.read(length)
				if not data or data.length < length
					raise Marathon::ExportError.new("Unexpected EOF reading chunk data")
				end
				write_chunk_header(f, tag, length, last)
				f.write(data)
				if last
					break
				end
			end
		rescue SystemCallError, Marathon::ExportError
			raise Marathon::ExportError.new(format("%s: %s", phys_path, $!.message))
		end
	end

	#---------------------------------------------------------------------------------

	def write_chunk(f, tag, data)
		bytes = data.join("")
		write_chunk_header(f, tag, bytes.length)
		f.write(bytes)
	end
	
	def read_chunk_header(f)
		return Marathon.read_map_chunk_header(f)
	end
	
	def write_chunk_header(f, tag, length, last = false)
		Marathon.write_map_chunk_header(f, 128, tag, length, last)
	end
	
	def write_struct(f, format, data, name, length)
		Marathon.write_struct(f, format, data, name, length)
	end

end # class MarathonExporter
