####################################################################
#
#   MarathUp - Level merging
#
####################################################################

module Marathon

	def Marathon.merge_levels()
		begin
			model = Sketchup.active_model
			if model
				dir = File.dirname(model.path)
			else
				dir = ""
			end
			level_list_path = UI.openpanel("Merge Levels - Select level list",
				dir, "Map.txt")
			if not level_list_path
				return
			end
			dir, level_list_name= File.split(level_list_path)
			map_path = UI.savepanel("Save merged levels as:",
				dir, replace_extension(level_list_name, ".sceA"))
			if not map_path
				return
			end
			map_dir, map_name = File.split(map_path)
			rsrc_path = find_resource_file(map_dir, replace_extension(map_name, ""))
			level_list = parse_level_list(level_list_path)
			if rsrc_path
				write_merged_macbinary_file(map_path, level_list, rsrc_path)
			else
				write_merged_plain_file(map_path, level_list)
			end
			UI.messagebox(format("Merged levels written to %s.", map_path))
		rescue Marathon::ExportError
			UI.beep
			UI.messagebox($!.message)
		end
	end
	
	def Marathon.replace_extension(name, ext)
		return name.sub(/\.[^.]*$/, "") + ext
	end
	
	def Marathon.write_merged_plain_file(map_path, level_list)
		f = File.open(map_path, "wb")
		write_merged_levels(f, map_path, level_list)
		f.close()
	end
	
	def Marathon.write_merged_macbinary_file(map_path, level_list, rsrc_path)
		f = MacBinaryFile.create(map_path, "sce2", "26.\xb0")
		f.open_data_fork()
		write_merged_levels(f, map_path, level_list)
		f.close_data_fork()
		copy_resource_fork(f, rsrc_path)
		f.close()
	end
	
	def Marathon.write_merged_levels(f, map_path, level_list)
		map_dir, map_name = File.split(map_path)
		write_map_header(f, map_name, level_list.length, true)
		for entry in level_list
			add_merged_level(f, map_dir, entry)
		end
		dir_offset = f.tell()
		printf("Writing directory at 0x%x\n", dir_offset) ###
		for entry in level_list
			printf("%s\n", entry.dump()) ###
			write_map_directory_entry(f,
				entry.offset, entry.length, entry.index - 1,
				[entry.mission_flags, entry.environment_flags,
					entry.entry_point_flags, entry.level_name])
		end
		fix_map_directory_offset(f, dir_offset)
	end
	
	def Marathon.add_merged_level(f, map_dir, entry)
		level_start = f.tell()
		level_path = File.join(map_dir, entry.filename)
		begin
			g = MacBinaryFile.open_data_fork(level_path)
		rescue Errno::ENOENT
			raise ExportError.new(format("Level file not found: %s", level_path))
		end
		printf("Skipping level file header\n") ###
		g.seek(128)
		last = false
		while not last
			start_offset = g.tell() - 128
			printf("Reading chunk at offset 0x%x\n", start_offset) ###
			tag, next_offset, length = read_map_chunk_header(g)
			printf("Chunk tag '%s' next_offset 0x%x length 0x%x\n",
				tag, next_offset, length) ###
			if next_offset == 0
				last = true
			elsif next_offset != start_offset + 16 + length
				raise ExportError.new(format("%s: 0x%x: Invalid chunk next_offset",
					level_path, start_offset + 128))
			end
			bytes = g.read(length)
			if bytes.length != length
				raise ExportError.new(format("%s: Unexpected end of file", level_path))
			end
			if tag == "Minf"
				if length != 88
					raise ExportError.new(format("%s: 0x%x: Wrong length Minf chunk",
						level_path, start_offset + 128))
				end
				data = bytes.unpack("n5a8a66N")
				entry.mission_flags = data[3]
				entry.environment_flags = data[4]
				entry.level_name = data[6]
				entry.entry_point_flags = data[7]
			end
			write_map_chunk_header(f, level_start, tag, length, last)
			f.write(bytes)
		end
		entry.offset = level_start
		entry.length = f.tell() - level_start
	end
	
	#------------------------------------------------------------------
	#
	#   Level list parsing
	#
	#------------------------------------------------------------------

	def Marathon.parse_level_list(path)
		f = File.open(path)
		lineno = 0
		entries = []
		while line = f.gets()
			lineno += 1
			index_str, filename = line.split()
			if index_str
				index = index_str.to_i()
				if index != entries.length + 1
					raise ExportError.new(format(
						"%s, line %s: Level number out of sequence", path, lineno))
				end
				if not filename
					raise ExportError.new(format("%s, line %s: Missing level file name",
						path, lineno))
				end
				entry = MapDirEntry.new(index, filename)
				entries.push(entry)
			end
		end
		return entries
	end
	
	#------------------------------------------------------------------

	class MapDirEntry
	
		attr :index
		attr :filename
		attr :mission_flags, true
		attr :environment_flags, true
		attr :level_name, true
		attr :entry_point_flags, true
		attr :offset, true
		attr :length, true
		
		def initialize(index, filename)
			@index = index
			@filename = filename
		end
		
		def dump()
			return format(
				"MapDirEntry %s %s '%s' miss 0x%x env 0x%x ep 0x%x off 0x%x len 0x%x",
				@index, @filename, @level_name,
				@mission_flags, @environment_flags, @entry_point_flags,
				@offset, @length)
		end
		
	end # class MapDirEntry

	#------------------------------------------------------------------

end # module Marathon

if not $marathon_merge_levels
	menu = marathon_submenu()
	menu.add_item("Merge Levels") {Marathon.merge_levels()}
	$marathon_merge_levels = true
end

