#################################################################
#
#   Sketchup to Marathon Converter - Monster and Item Import
#
#   Imports a set of image files containg front and back
#   views of monsters or items and creates component definitions
#   from them.
#
#   Names of the component definitions are taken from a text
#   file containing lines of the form
#
#      <index>-<name>
#
#   where <index> is the monster number. The index can be
#   followed by an optional letter suffix 'a', 'b', etc.
#   to accommodate different monsters sharing the same
#   monster slot.
#
#   The images should be in a folder with the same name as
#   the names file with the extension removed.
#
#   The image files should be named
#
#      <index>-<filename>-Front.<ext>
#      <index>-<filename>-Back.<ext>
#
#   where <index> includes the letter suffix, if any, and
#   <ext> is a know image file type. The full image file
#   name is used as the name of the imported texture.
#
#################################################################

module Marathon

	class ComponentImporter
	
		YES = 6
		NO = 7
		CANCEL = 2
	
		attr :errors
	
		def run(type, namefile_path)
			begin
				@errors = []
				@images = {}
				@type = type
				self.find_image_dir(namefile_path)
				self.read_namefile(namefile_path)
				#self.dump_name_table()
				self.find_images()
				#self.dump_image_table()
				self.load_textures()
				self.create_components()
				return @errors.length == 0
			rescue SystemCallError
				UI.messagebox($!.message)
				throw :Abort
			end
		end
		
		def find_image_dir(namefile_path)
			@dir = remove_extension(namefile_path)
			begin
				kind = File.ftype(@dir)
			rescue IOError
				kind = ""
			end
			if kind != "directory"
				self.error("Directory %s not found", @dir)
				throw :Abort
			end
		end
		
		def remove_extension(path)
			return path.sub(/\.[^.]+$/, "")
		end
			
		NamefileLinePat = /^([0-9]+)([a-z]*)-(.+)$/
		
		def read_namefile(path)
			@names = {}
			@keys = {}
			f = File.open(path)
			lineno = 0
			while line = f.gets()
				lineno += 1
				line = line.strip()
				if line != ""
					match = NamefileLinePat.match(line)
					if not match
						self.error("%s, line %s: Invalid name specification", path, lineno)
						next
					end
					index, suffix, name = match[1, 3]
					self.add_name(index, suffix, name)
				end
			end
			f.close()
			if @errors.length > 0
				throw :Abort
			end
		end
		
		def add_name(index, suffix, name)
			key = [index, suffix]
			if @keys[name]
				self.error("%s, line %s: Multiple entries named '%s'",
					path, lineno, name)
			end
			@keys[name] = key
			if not @names[key]
				@names[key] = name
			else
				self.error("%s, line %s: Multiple names for index %s%s",
					path, lineno, index, suffix)
			end
		end
		
		def find_images()
			pat = File.join(@dir, "*.*")
			image_paths = Dir.glob(pat)
			image_paths.sort!()
			for path in image_paths
				self.add_image(path)
			end
		end
		
		ImageNamePat = /^([0-9]+)([a-z]*)-[^-]*-([^-]+)\./
		
		def add_image(path)
			name = File.basename(path)
			match = ImageNamePat.match(name)
			if not match
				self.error("%s: Invalid image file name", path)
				return
			end
			index, suffix, face = match[1, 3]
			if not ["Front", "Back"].index(face)
				self.error("%s: Invalid face '%s'", path, face)
				return
			end
			key = [index, suffix]
			entry = @images[key]
			if not entry
				entry = {}
				@images[key] = entry
			end
			if entry[face]
				self.error("%s: Multiple images for index %s%s face %s",
					path, index, suffix, face)
				return
			end
			entry[face] = path
		end
		
		def load_textures()
			materials = Sketchup.active_model.materials
			for key, entry in @images
				for face, path in entry
					name = texture_name(path)
					#printf("Processing texture %s\n", name.inspect) ###
					material = materials[name]
					if material and not self.reloading
						next
					end
					if not material
						printf("Adding texture %s\n", name.inspect) ###
						material = materials.add(name)
					end
					printf("Loading texture %s\n", name.inspect) ###
					material.texture = path
					texture = material.texture
					width = 0.75 * texture.image_width / 128.0
					height = 0.75 * texture.image_height / 128.0
					if face == "Back"
						width = -width
					end
					texture.size = [width, height]
				end
			end
		end
		
		def texture_name(path)
			if not path
				return nil
			end
			return remove_extension(File.basename(path))
		end
		
		def reloading()
			if @reloading == nil
				ans = UI.messagebox("Reload existing textures?", MB_YESNOCANCEL)
				if ans == CANCEL
					throw :Abort
				end
				@reloading = (ans == YES)
			end
			return @reloading
		end
				
		def updating()
			if @updating == nil
				ans = UI.messagebox("Update existing components?", MB_YESNOCANCEL)
				if ans == CANCEL
					throw :Abort
				end
				@updating = (ans == YES)
			end
			return @updating
		end
				
		def create_components()
			for key, component_name in @names
				self.create_component(key, component_name)
			end
		end
		
		def create_component(key, component_name)
			model = Sketchup.active_model
			materials = model.materials
			definitions = model.definitions
			defn = definitions[component_name]
			if defn and not self.updating
				return
			end
			#printf("Processing component %s %s\n", key.inspect, component_name.inspect)
			entry = @images[key]
			if not entry
				self.error("No images for %s: '%s'", key, component_name)
				return
			end
			front_material = materials[texture_name(entry["Front"])]
			back_texture_name = texture_name(entry["Back"])
			if back_texture_name
				back_material = materials[]
			else
				back_material = nil
			end
			if not front_material
				self.error("No Front image for %s: %s", key, component_name)
				return
			end
			front_texture = front_material.texture
			if not front_texture
				self.error("Material '%s' is not a texture", front_material.name)
				return
			end
			if not defn
				index = key[0]
				defn = definitions.add(component_name)
				defn.description = format("*%s,%s", @type, index)
			end
			ents = defn.entities
			ents.clear!()
			w = front_texture.width
			h = front_texture.height
			face = ents.add_face([[0, 0, 0], [0, w, 0], [0, w, h], [0, 0, h]])
			edges = face.edges
			for i in 1..3
				edges[i].hidden = true
			end
			face.material = front_material
			face.back_material = back_material or front_material
		end
		
		def error(fmt, *args)
			@errors.push(format(fmt, *args))
		end
		
		def dump()
			self.dump_name_table()
			self.dump_image_table()
		end
		
		def dump_name_table()
			printf("Name table:\n")
			for key in @names.keys.sort()
				printf("  %s => %s\n", key.inspect, @names[key].inspect)
			end
		end
		
		def dump_image_table()
			printf("Image table:\n")
			for key in @images.keys.sort()
				entry = @images[key]
				printf("  %s:\n", key.inspect)
				for face, path in entry
					printf("    %s => %s\n", face.inspect, path.inspect)
				end
			end
		end
	
	end # class ComponentImporter
	
	def Marathon.import_components(type)
		catch :Abort do
			prompt = format("Import %ss: Select Name File", type)
			model = Sketchup.active_model
			dir = Sketchup.read_default("Marathon", "MonsterImportDirectory", "")
			if dir == ""
				dir = File.dirname(model.path)
			end
			default_name = format("%ss.txt", type)
			path = UI.openpanel(prompt, dir, default_name)
			if path and path != ""
				dir = File.dirname(path)
				Sketchup.write_default("Marathon", "MonsterImportDirectory", dir)
				importer = ComponentImporter.new()
				if importer.run(type, path)
					#importer.dump() ###
					UI.messagebox(format("%s imported successfully.", type))
				else
					UI.beep
					UI.messagebox(importer.errors.join("\n"), MB_MULTILINE,
						format("%s Import Messages", type))
				end
			end
		end
	end
	
	def Marathon.import_monsters()
		import_components("Monster")
	end
	
	def Marathon.import_items()
		import_components("Item")
	end
	
	def Marathon.import_scenery()
		import_components("Scenery")
	end
	
end # module Marathon

if (not $marathon_import_monsters)
	menu = marathon_submenu()
	menu.add_item("Import Monsters") {Marathon.import_monsters()}
	$marathon_import_monsters = true
end

if (not $marathon_import_items)
	menu = marathon_submenu()
	menu.add_item("Import Items") {Marathon.import_items()}
	$marathon_import_items = true
end

if (not $marathon_import_scenery)
	menu = marathon_submenu()
	menu.add_item("Import Scenery") {Marathon.import_scenery()}
	$marathon_import_scenery = true
end
