##################################################################################
#
#   Sketchup to Marathon Export Script
#
##################################################################################

load("Marathon/Export/MacBinary.rb")
load("Marathon/Export/DataStructures.rb")
load("Marathon/Export/Terminal.rb")
load("Marathon/Export/MapFileOutput.rb")

def marathon_coord(x)
	return (x * 1024).round
end

def marathon_coords(coords)
	return coords.map {|x| (x * 1024).round}
end

def marathon_speed(wu_per_sec)
	return (wu_per_sec * 1024 / 30.0).round()
end

module Marathon

	class ExportError < StandardError
	end # class ExportError

end # module Marathon
	
class MarathonExporter

	@@Debug = false
	@@DebugSides = false
	@@DebugControlPanels = false
	@@DebugLights = false
	@@DebugMedia = false
	@@DebugObjects = false
	@@VisibleOnly = false
	
	def MarathonExporter.DebugLights
		return @@DebugLights
	end

	RIGHT = 0
	LEFT = 1
	
	BOTTOM = 0
	TOP = 1
	
	FLOOR = 0
	CEILING = 1
	
	CLOCKWISE = RIGHT
	ANTICLOCKWISE = LEFT
	
	MarathonObjectTypes = {
		"*Monster" => 0,
		"*Scenery" => 1,
		"*Item" => 2,
		"*Player" => 3,
		"*Goal" => 4,
		"*Sound" => 5,
	}
	
	DefaultLightAttrs = {
		"flags" => 0x1, # _light_is_initially_active
		"becoming_active.function" => 2,
		"becoming_active.period" => 30,
		"becoming_active.intensity" => 1.0,
		"primary_active.function" => 0,
		"primary_active.period" => 30,
		"primary_active.intensity" => 1.0,
		"secondary_active.function" => 0,
		"secondary_active.period" => 30,
		"secondary_active.intensity" => 1.0,
		"becoming_inactive.function" => 2,
		"becoming_inactive.period" => 30,
		"primary_inactive.function" => 0,
		"primary_inactive.period" => 30,
		"secondary_inactive.function" => 0,
		"secondary_inactive.period" => 30,
	}

	#---------------------------------------------------------------------------------
	#
	#   MarathonExporter Initialization
	#
	#---------------------------------------------------------------------------------

	def initialize()
		@walls = []
		@floors = []
		@ceilings = []
		@objects = []                   # List of [instance, transformation]
		@lights = []										# List of attribute dict
		@media = []                     # List of [instance, transformation]
		@random_sounds = []             # List of [instance, transformation]
		@ambient_sounds = []            # List of [instance, transformation]
		@notes = []                     # List of [instance, transformation]
		@marathon_points = []           # List of [x,y]
		@marathon_polys = []            # List of MarathonPoly
		@marathon_lines = []            # List of MarathonLine
		@marathon_sides = []            # List of MarathonSide
		@marathon_objects = []          # List of MarathonObject
		@marathon_platforms = []        # List of MarathonPlatform
		@marathon_terminals = []        # List of MarathonTerminal
		@coords_to_marathon_point = {}  # [x,y] --> point number
		@corners_to_marathon_polys = ListMap.new() # Sorted point nos. to list of MarathonPoly
		@edge_to_marathon_polys = ListMap.new()    # [Sorted point nos, z] to list of [MarathonPoly, edge no.]
		@item_counts = Array.new(64, 0)
		@monster_counts = Array.new(64, 0)
		@number_of_tags = 0
		@number_of_goals = 0
		@name_to_marathon_poly = {}
		@name_to_tag_number = {}
		@name_to_goal_number = {}
		@name_to_light_number = {}
		@name_to_media_number = {}
		@name_to_ambient_sound_number = {}
		@name_to_random_sound_number = {}
		@name_to_terminal = {}
		@name_to_marathon_object = {}
		self.add_light("", DefaultLightAttrs)
	end

	#---------------------------------------------------------------------------------
	#
	#   Debugging
	#
	#---------------------------------------------------------------------------------

	def dump_marathon_points()
		print("Marathon Points:\n")
		for i in 0...@marathon_points.length
			p = @marathon_points[i]
			printf("%d: %s\n", i, p.inspect)
		end
	end
	
	def dump_marathon_polys()
		print("Marathon Polys:\n")
		for poly in @marathon_polys
			printf("%d: %s\n", poly.number, poly.dump)
		end
		print("Marathon Polys by Corners:\n")
		for key, polys in @corners_to_marathon_polys
			val = polys.map {|poly| poly.number}
			printf("%s --> %s\n", key.inspect, val.inspect)
		end
		print("Marathon Polys by Edge:\n")
		for key, polys in @edge_to_marathon_polys
			val = polys #.map {|poly| poly.number}
			printf("%s --> %s\n", key.inspect, val.inspect)
		end
	end
	
	def dump_marathon_lines()
		print("Marathon Lines:\n")
		for line in @marathon_lines
			printf("%d: %s\n", line.number, line.dump)
		end
	end
	
	def dump_marathon_sides()
		print("Marathon Sides:\n")
		for side in @marathon_sides
			printf("%d: %s\n", side.number, side.dump)
		end
	end
	
	def dump_marathon_objects()
		print("Marathon Objects:\n")
		for obj in @marathon_objects
			printf("%d: %s\n", obj.number, obj.dump)
		end
	end
	
	def dump_counts(name, counts)
		printf("%s counts: %s\n", name, counts.inspect)
	end
	
	def dump()
		if @@Debug
			self.dump_marathon_points()
			self.dump_marathon_polys()
			self.dump_marathon_lines()
			self.dump_marathon_sides()
			self.dump_marathon_objects()
			self.dump_counts("Item", @item_counts)
			self.dump_counts("Monster", @monster_counts)
		end
	end
	
	#---------------------------------------------------------------------------------
	#
	#   Error Reporting
	#
	#---------------------------------------------------------------------------------

	def raise_error(mess)
		raise Marathon::ExportError.new(mess)
	end

	def error(obj, ctx, mess)
		p = ctx.trans * obj.bounds.center
		error_at_point3d(p, mess)
	end
	
	def edge_error(edge, mess)
		(i, j), z = edge
		p0 = self.marathon_to_point3d(i, z)
		p1 = self.marathon_to_point3d(j, z)
		p = Geom::Point3d.linear_combination(0.5, p0, 0.5, p1)
		error_at_point3d(p, mess)
	end
	
	def marathon_to_point3d(p, z)
		x, y = @marathon_points[p]
		return Geom::Point3d.new([x / 1024.0, y / 1024.0, z / 1024.0])
	end
	
	def error_at_point3d(p, mess)
		v = Geom::Vector3d.new([0.25, 0.25, 0.25])
		text = Sketchup.active_model.entities.add_text(mess, p, v)
		text.leader_type = 1
		text.layer = self.error_layer()
		UI.beep
		UI.messagebox(mess)
		throw :MarathonAbort
	end
	
	def error_layer()
		name = "*Marathon Errors*"
		layers = Sketchup.active_model.layers
		return layers.add(name)
	end
	
	def delete_error_markers()
		el = self.error_layer()
		markers = []
		for entity in Sketchup.active_model.entities
			if entity.layer == el
				markers.push entity
			end
		end
		for marker in markers
			marker.erase!
		end
	end
	
#	def show_only_layer(l)
#		model = Sketchup.active_model
#		model.active_layer = l
#		for layer in model.layers
#			layer.visible = (layer == l)
#		end
#	end
	
	#---------------------------------------------------------------------------------
	#
	#   Utilities
	#
	#---------------------------------------------------------------------------------

	def new_marathon_poly(name, points)
		n = @marathon_polys.length
		poly = MarathonPoly.new(n, points)
		@marathon_polys.push(poly)
		if name != ""
			@name_to_marathon_poly[name] = poly
		end
		return poly
	end
	
	def new_marathon_line(points)
		n = @marathon_lines.length
		line = MarathonLine.new(n, points)
		@marathon_lines.push(line)
		return line
	end
	
	def new_marathon_object()
		n = @marathon_objects.length
		obj = MarathonObject.new(n)
		@marathon_objects.push(obj)
		return obj
	end
	
	def new_marathon_platform(poly)
		n = @marathon_platforms.length
		plat = MarathonPlatform.new(n, poly)
		@marathon_platforms.push(plat)
		poly.platform = plat
		return plat
	end
	
	def marathon_point_number(coords)
		if @coords_to_marathon_point.has_key?(coords)
			return @coords_to_marathon_point[coords]
		else
			i = @marathon_points.length
			@marathon_points.push(coords)
			@coords_to_marathon_point[coords] = i
			return i
		end
	end

	def corner_point_numbers_for_face(face, ctx)
		trans = ctx.trans
		result = []
		for v in face.vertices
			p = trans * v.position
			xy = marathon_coords([p.x, p.y])
			i = self.marathon_point_number(xy)
			result.push(i)
		end
		return result
	end
	
	def number(obj)
		if obj
			return obj.number
		else
			return -1
		end
	end
	
	def numbers(objs)
		return objs.map {|obj| number(obj)}
	end
	
	def marathon_shape_descriptor(material)
		if material
			return MarathonShapeDescriptor.from_name(material.name)
		end
	end

	def poly_named(name)
		#  Find polygon with specified name. If name is empty, returns
		#  nil. Otherwise, raises an error if the polygon doesn't exist.
		if name == nil or name == ""
			return nil
		end
		poly = @name_to_marathon_poly[name]
		if not poly
			raise_error(format('No polygon named "%s".', name))
		end
		return poly
	end
	
	def require_poly_named(name)
		#  Find polygon with specified name. Raises an error if the
		#  name is empty or the polygon doesn't exist.
		if name == nil or name == ""
			raise_error("Polygon name required.")
		end
		return self.poly_named(name)
	end

	def number_of_light_named(name)
		#  Return the light number of the light with the specified name.
		#  If name is empty, returns the default light number. Otherwise,
		#  raises an error if the light does not exist.
		if name == nil
			name = ""
		end
		n = @name_to_light_number[name]
		if n == nil
			raise_error(format('No light named "%s".', name))
		end
		return n
	end
	
	def number_of_media_named(name)
		#  Return the media number of the media with the specified name.
		#  If name is empty, returns -1. Otherwise, raises an error if the
		#  media does not exist.
		if name == nil or name == ""
			return -1
		end
		n = @name_to_media_number[name]
		if n == nil
			raise_error(format('No liquid named "%s".', name))
		end
		return n
	end
	
	def number_of_ambient_sound_named(name)
		#  Return the ambient sound number of the ambient sound with the
		#  specified name. If name is empty, returns -1. Otherwise, raises
		#  an error if the ambient sound does not exist.
		if name == nil or name == ""
			return -1
		end
		n = @name_to_ambient_sound_number[name]
		if n == nil
			raise_error(format('No ambient sound named "%s".', name))
		end
		return n
	end
	
	def number_of_random_sound_named(name)
		#  Return the random sound number of the random sound with the
		#  specified name. If name is empty, returns -1. Otherwise, raises
		#  an error if the random sound does not exist.
		if name == nil or name == ""
			return -1
		end
		n = @name_to_random_sound_number[name]
		if n == nil
			raise_error(format('No random sound named "%s".', name))
		end
		return n
	end
	
	def tag_named(name)
		#  Return the tag number corresponding to the name. If the name
		#  is empty, returns 0. Otherwise, a new tag number > 0 is allocated
		#  if the name has not been seen before.
		if name == nil or name == ""
			return 0
		end
		n = @name_to_tag_number[name]
		if n == nil
			n = @number_of_tags + 1
			@number_of_tags = n
			@name_to_tag_number[name] = n
		end
		return n
	end

	def goal_named(name)
		#  Return the goal number corresponding to the name. If the name
		#  is empty, returns -1. Otherwise, a new goal number >= 0 is allocated
		#  if the name has not been seen before.
		if name == nil or name == ""
			return -1
		end
		n = @name_to_goal_number[name]
		if n == nil
			n = @number_of_goals
			@number_of_goals = n + 1
			@name_to_goal_number[name] = n
		end
		return n
	end

	def terminal_named(name)
		#  Return the terminal corresponding to the given filename. If the
		#  name is empty, returns nil. Otherwise, the terminal is assumed
		#  to exist.
		if name == nil or name == ""
			return nil
		end
		return @name_to_terminal[name]
	end
	
	def object_named(name)
		#  Returns the object with the given name. If the name is empty,
		#  returns -1. Otherwise, raises an error if the object does not
		#  exist.
		if name == nil or name == ""
			return -1
		end
		obj = @name_to_marathon_object[name]
		if not obj
			raise_error(format("No object named '%s'", name))
		end
		return obj
	end

	#---------------------------------------------------------------------------------
	#
	#   Polygon Creation
	#
	#   Pair up matching floor and ceiling faces and create polygons from them.
	#
	#---------------------------------------------------------------------------------

	def create_marathon_polys()
		self.create_marathon_polys_from_floors()
		self.attach_ceilings_to_marathon_polys()
		self.check_for_nonconvex_polys()
		self.index_marathon_polys_by_edge()
	end

	def create_marathon_polys_from_floors()
		for face, ctx in @floors
			begin
				self.create_marathon_poly_from_floor(face, ctx)
			rescue Marathon::ExportError
				error(face, ctx, $!.message)
			end
		end
	end
	
	def create_marathon_poly_from_floor(face, ctx)
		trans = ctx.trans
		p = trans * face.vertices[0].position
		z = marathon_coord(p.z)
		pts = self.corner_point_numbers_for_face(face, ctx)
		#  A mirrored floor face needs the order of its points reversed
		#  so that they remain clockwise.
		if is_mirrored(trans)
			pts.reverse!()
		end
		a = (face.area * 1024 * 1024).round
		cp = face.bounds.center
		cxy = marathon_coords([cp.x, cp.y])
		poly_attrs = Marathon.get_polygon_attributes(face)
		floor_attrs = Marathon.get_floor_ceiling_attributes(face)
		name = (poly_attrs["name"] or "")
		poly = self.new_marathon_poly(name, pts)
		poly.floor = face
		poly.floor_ctx = ctx
		poly.floor_height = z
		poly.area = a
		poly.center = cxy
		poly.floor_texture = marathon_shape_descriptor(face.back_material)
		poly.floor_origin = calculate_texture_origin(face, trans)
		poly.floor_light = ctx.eval_param(floor_attrs["light"])
		poly.floor_transfer_mode = (floor_attrs["transfer_mode"] or 0)
		poly.permutation_name = ctx.eval_param(poly_attrs["permutation"])
		poly.media_name = ctx.eval_param(poly_attrs["media_index"])
		poly.media_light = ctx.eval_param(poly_attrs["media_lightsource_index"])
		poly.ambient_sound = ctx.eval_param(poly_attrs["ambient_sound_image_index"])
		poly.random_sound = ctx.eval_param(poly_attrs["random_sound_image_index"])
		key = pts.sort()
		@corners_to_marathon_polys[key].push(poly)
	end

	def attach_ceilings_to_marathon_polys()
		for face, ctx in @ceilings
			begin
				self.attach_ceiling_to_marathon_poly(face, ctx)
			rescue Marathon::ExportError
				error(face, ctx, $!.message)
			end
		end
	end

	def attach_ceiling_to_marathon_poly(face, ctx)
		trans = ctx.trans
		p = trans * face.vertices[0].position
		z = marathon_coord(p.z)
		key = self.corner_point_numbers_for_face(face, ctx).sort()
		poly = nil
		for candidate in @corners_to_marathon_polys[key]
			if candidate.floor_height <= z
				if poly == nil or poly.floor_height < candidate.floor_height
					poly = candidate
				end
			end
		end
		if poly == nil
			error(face, ctx, "Unable to find floor corresponding to ceiling face.")
		end
		attrs = Marathon.get_floor_ceiling_attributes(face)
		poly.ceiling = face
		poly.ceiling_ctx = ctx
		poly.ceiling_height = z
		poly.ceiling_texture = marathon_shape_descriptor(face.back_material)
		poly.ceiling_origin = calculate_texture_origin(face, trans)
		poly.ceiling_light = ctx.eval_param(attrs["light"])
		poly.ceiling_transfer_mode = (attrs["transfer_mode"] or 0)
	end

	def is_mirrored(trans)
		return (trans.xaxis * trans.yaxis) % trans.zaxis < 0
	end
	
	def calculate_texture_origin(face, trans)
		#  Calculate origin of floor or ceiling texture.
		x0 = y0 = 0
		attrs = Marathon.get_floor_ceiling_attributes(face)
		if attrs["align_texture"]
			material = face.back_material
			if material
				mesh = face.mesh(2) # back uvs
				p = trans * mesh.point_at(1)
				uv = mesh.uv_at(1, false) # back uv
				#printf("F/C point %s has uv %s\n", p.inspect, uv.inspect) ###
				x0 = -marathon_coord(p.x + uv.y) & 0x3ff
				y0 = -marathon_coord(p.y + uv.x) & 0x3ff
			end
		end
		#printf("x0 = %s, y0 = %s\n", x0, y0) ###
		return [x0, y0]
	end
	
	def poly_point(poly, i)
		# Coordinates of vertex i mod number of points
		return @marathon_points[poly.point_number(i)]
	end

	def check_for_nonconvex_polys()
		for poly in @marathon_polys
			n = poly.num_points
			for i in 0...n
				x0, y0 = self.poly_point(poly, i)		
				x1, y1 = self.poly_point(poly, i + 1)		
				x2, y2 = self.poly_point(poly, i + 2)
				d = (x1 - x0) * (y2 - y1) - (x2 - x1) * (y1 - y0)
				if d > 0
					error(poly.ceiling, poly.ceiling_ctx,
						"Polygon is non-convex.")
				end
			end
		end
	end
	
	def index_marathon_polys_by_edge()
		for poly in @marathon_polys
			if poly.ceiling == nil
				error(poly.floor, poly.floor_ctx,
					"Unable to find ceiling corresponding to floor face.")
			end
			for i in 0...poly.num_points
				heights = poly.heights
				pts = poly.edge_point_numbers(i).sort()
				#printf("Adding edge-to-poly mapping for Poly %d points %s heights %s\n",
				#	poly.number, pts.inspect, heights.inspect) ###
				for z in heights
					key = [pts, z]
					@edge_to_marathon_polys[key].push([poly, i])
				end
			end
		end
	end

	#---------------------------------------------------------------------------------
	#
	#   Line Creation, Part 1
	#
	#   Create Marathon Lines corresponding to wall faces and connect them to
	#   adjacent Polygons.
	#
	#---------------------------------------------------------------------------------

	def create_marathon_lines_from_walls()
		#print("Creating lines from walls\n") ###
		for face, ctx in @walls
			wall = self.create_wall_descriptor_from_face(face, ctx)
			connects_to_ceiling = false
			connects_to_floor = false
			polys = {} # MarathonPoly to edge no. of poly
			for key in wall.edges
				z = key[1]
				for poly, i in @edge_to_marathon_polys[key]
					#printf("Wall %s edge %s connects with poly %d ceiling %s floor %s\n",
					#	face, key.inspect, poly.number, poly.ceiling_height, poly.floor_height) ###
					if not polys.has_key?(poly)
						polys[poly] = i
					end
					if z == poly.ceiling_height
						connects_to_ceiling = true
					end
					if z == poly.floor_height
						connects_to_floor = true
					end
				end
			end
			self.check_number_of_polys_connected_to_wall(polys.size, face, ctx)
			line = nil
			for poly, i in polys
				if line == nil
					line = poly.lines[i]
				end
			end
			if line == nil
				line = self.new_marathon_line(wall.point_numbers)
			end
			for poly, i in polys
				if poly.lines[i] == nil
					line.connect_to_poly(poly, i)
				end
			end
			if connects_to_ceiling and connects_to_floor
				line.spanning_wall = wall
			elsif connects_to_ceiling
				line.upper_wall = wall
			elsif connects_to_floor
				line.lower_wall = wall
			end
		end
	end

	def check_number_of_polys_connected_to_wall(n, face, ctx)
		if n < 1
			error(face, ctx, "Wall face does not connect to any floor or ceiling faces.")
		end
		if n > 2
			error(face, ctx, "Wall face connects to more than 2 floor or ceiling faces.")
		end
	end

	def create_wall_descriptor_from_face(face, ctx)
		#  Build a wall descriptor from a wall face. The wall descriptor holds
		#  the wall's endpoint numbers and heights and information about the
		#  texturing applied to each side.
		trans = ctx.trans
		mirrored = is_mirrored(trans)
		(i, j), z0, z1 = extract_edges_from_wall_face(face, ctx, mirrored)
		materials = [face.material, face.back_material]
		uvs = extract_uvs_from_face(face, trans)
		if false ###
			printf("Creating wall descriptor from %s\n", face) ###
			mat = materials[0] != nil || materials[1] != nil ###
			if mat ###
				printf("points = %s heights = %s, %s\n", [i, j].inspect, z0, z1) ###
				printf("materials = %s\n", materials.inspect) ###
				printf("uvs = %s\n", uvs.inspect) ###
			end ###
		end ###
		if i > j # then swap front and back
			i, j = j, i
			uvs = [uvs[1], uvs[0]]
			materials = [materials[1], materials[0]]
		end
		# materials and uvs are now ordered right, left
		e = [i, j]
		wall = WallDescriptor.new()
		wall.face = face
		wall.ctx = ctx
		wall.point_numbers = e
		wall.heights[BOTTOM], wall.heights[TOP] = z0, z1
		sides = []
		for s in 0..1
			material = materials[s]
			if material
				side = WallSide.new()
				side.face = face
				side.ctx = ctx
				side.material = material
				uv = uvs[s]
				#  If the face is mirrored, use uvs from the right hand edge instead of
				#  the left. This results in the correct portion of the texture being
				#  shown at play time, although it will appear reversed in Sketchup.
				if mirrored
					k = 1 - s
				else
					k = s
				end
				bottom_uv = uv[[e[k], z0]]
				top_uv = uv[[e[k], z1]]
				side.uvs[BOTTOM] = bottom_uv
				side.uvs[TOP] = top_uv
			else
				side = nil
			end
			sides.push(side)
		end
		wall.sides[RIGHT], wall.sides[LEFT] = sides
		return wall
	end

	def extract_uvs_from_face(face, trans)
		#  Given a face, return two mappings [front, back] from [i, z] to uv
		#  where i is a point number and z a height.
		mat = false ###
		#mat = (face.material or face.back_material) ###
		if mat ###
			printf("Extracting UVs from %s\n", face) ###
		end ###
		mesh = face.mesh(3) # Include front and back UVs
		pts = mesh.points
		front_uvs = mesh.uvs(true)
		back_uvs = mesh.uvs(false)
		if mat ###
			printf("front %s back %s\n", front_uvs, back_uvs) ###
		end if ###
		uvs = [{}, {}]
		for i in 0...pts.length
			p = trans * pts[i]
			j = self.marathon_point_number(marathon_coords([p.x, p.y]))
			z = marathon_coord(p.z)
			k = [j, z]
			uvs[0][k] = front_uvs[i]
			uvs[1][k] = back_uvs[i]
		end
		if mat ###
			printf("result %s\n", uvs.inspect) ###
		end ###
		return uvs
	end
	
	def extract_edges_from_wall_face(face, ctx, mirrored)
		#  Given a wall face, return [bottom_edge, top_edge] where each edge
		#  is [[i, j], z] where i, j are point numbers with i on the left
		#  and j on the right when looking at the front of the face.
		xy = []
		z = []
		trans = ctx.trans
		for v in face.outer_loop.vertices
			p = trans * v.position
			xy.push(marathon_coords([p.x, p.y]))
			z.push(marathon_coord(p.z))
		end
		if mirrored
			xy.reverse!()
			z.reverse!()
		end
		list = []
		n = xy.length
		for k in 0...xy.length
			l = (k + 1) % n
			xy0, xy1 = xy[k], xy[l]
			z0, z1 = z[k], z[l]
			if xy0 != xy1
				if z0 != z1
					error(face, ctx, "Wall edge not horizontal or vertical.")
				end
				i = self.marathon_point_number(xy0)
				j = self.marathon_point_number(xy1)
				list.push([z0, [i, j]])
			end
		end
		check_number_of_horizontal_wall_edges(list.length, face, ctx)
		list.sort!()
		(z0, (i0, j0)), (z1, (i1, j1)) = list
		if z0 == z1
			error(face, ctx, "Zero-height wall face.")
		end
		return [[i0, j0], z0, z1]
	end
	
	def check_number_of_horizontal_wall_edges(n, face, ctx)
		if n < 2
			error(face, ctx, "Wall face has less than 2 horizontal edges.")
		end
		if n > 2
			error(face, ctx, "Wall face has more than 2 horizontal edges.")
		end
	end

	#---------------------------------------------------------------------------------
	#
	#   Line Creation, Part 2
	#
	#   Create Lines where two Polygons connect without any intervening wall faces
	#   (i.e. the ceilings and floors are at the same heights). Check that no poly
	#   edge remains without a Line attached.
	#
	#---------------------------------------------------------------------------------

	def create_marathon_lines_without_walls()
		for poly in @marathon_polys
			for i in 0...poly.num_points
				if not poly.lines[i]
					ep = poly.edge_point_numbers(i).sort()
					z0 = poly.floor_height
					z1 = poly.ceiling_height
					adj_poly, j = self.find_adjacent_poly(poly, [ep, z0], FLOOR)
					adj_poly2, k = self.find_adjacent_poly(poly, [ep, z1], CEILING)
					if not adj_poly
						self.edge_error([ep, z0],
							"Floor edge not connected to a wall or another floor.")
					end
					if not adj_poly2
						self.edge_error([ep, z1],
							"Ceiling edge not connected to a wall or another ceiling.")
					end
					if adj_poly != adj_poly2
						error(poly.ceiling, poly.ceiling_ctx,
							"Poly has floor and ceiling connected to different polys.")
					end
					if j != k
						raise "Disagreement over floor/ceiling edge numbers of adjacent poly."
					end
					line = adj_poly.lines[j]
					if not line
						line = new_marathon_line(ep)
						line.connect_to_poly(adj_poly, j)
					end
					line.connect_to_poly(poly, i)
				end
			end
		end
	end
	
	def find_adjacent_poly(poly, edge, which_height)
		z = edge[1]
		result = nil
		for candidate, edge_no in @edge_to_marathon_polys[edge]
			if candidate != poly and candidate.heights[which_height] == z
				if result == nil
					result = [candidate, edge_no]
				else
					self.edge_error(edge,
						"More than 2 floor or ceiling faces connected at edge.")
				end
			end
		end
		return result
	end

	#---------------------------------------------------------------------------------
	#
	#   Side Creation
	#
	#   Create Marathon Sides from Marathon Lines and wall descriptors. Also
	#   calculates Line flags from adjacency information and user settings.
	#
	#---------------------------------------------------------------------------------

	def create_marathon_sides()
		for line in @marathon_lines
			begin
				self.create_sides_for_marathon_line(line)
			rescue Marathon::ExportError
				wall = line.representative_wall
				error(wall.face, wall.ctx, $!.message)
			end
		end
	end
	
	def create_sides_for_marathon_line(line)
		cwp = line.polygons[CLOCKWISE]
		acwp = line.polygons[ANTICLOCKWISE]
		span = line.spanning_wall
		exterior = (cwp == nil or acwp == nil)
		marked_solid = false
		marked_opaque = false
		face_flags = line.merged_wall_line_flags()
		if face_flags & 0x4000 != 0 # SOLID_LINE_BIT
			#printf("Line %s marked solid\n", line.inspect) ###
			marked_solid = true
		end
		if face_flags & 0x2000 != 0
			#printf("Line %s marked opaque\n", line.inspect) ###
			marked_opaque = true
		end
		flags = 0
		if exterior or marked_solid
			flags |= 0x4000 # SOLID_LINE_BIT
		end
		if not exterior and not marked_opaque
			flags |= 0x2000 # TRANSPARENT_LINE_BIT
		end
		line.flags = flags
		if cwp != nil
			self.create_marathon_side(line, RIGHT, exterior)
		end
		if acwp != nil
			self.create_marathon_side(line, LEFT, exterior)
		end
	end
	
	def create_marathon_side(line, which_side, exterior)
		upper_wall = line.upper_wall_side(which_side)
		lower_wall = line.lower_wall_side(which_side)
		spanning_wall = line.spanning_wall_side(which_side)
		attr_wall = (spanning_wall or lower_wall or upper_wall)
		if not attr_wall
			return
		end
		side = MarathonSide.new(@marathon_sides.length)
		@marathon_sides.push(side)
		if @@DebugSides
			printf("----- Creating side %s\n", side.number)
			printf("Upper wall = %s\n", upper_wall.inspect)
			printf("Lower wall = %s\n", lower_wall.inspect)
			printf("Spanning wall = %s\n", spanning_wall.inspect)
			printf("Attr wall = %s\n", attr_wall.inspect)
		end
		is_control_panel = false
		dual_texture = false
		side_flags = 0
		panel_type = -1
		if attr_wall
			ctx = attr_wall.ctx
			attrs = Marathon.get_wall_attributes(attr_wall.face)
			side_flags = (attrs["side_flags"] or 0)
			panel_type = (attrs["control_panel_type"] or -1)
			if panel_type >= 0
				if @@DebugControlPanels
					printf("Control panel type %s\n", panel_type)
				end
				is_control_panel = true
			end
		else
			attrs = {}
		end
		side.attrs = attrs
		if (upper_wall or lower_wall) and (exterior) ### or is_control_panel)
			wall = line.representative_wall
			#if exterior
				message = "Exterior wall cannot be a split wall."
			#else
			#	message = "Split wall cannot be a control panel."
			#end
			error(wall.face, wall.ctx, message)
		end
		upper_texture = texture_definition(upper_wall, TOP) #BOTTOM)
		lower_texture = texture_definition(lower_wall, TOP)
		spanning_texture = texture_definition(spanning_wall, TOP)
		side.primary_texture = NoTexture
		side.secondary_texture = NoTexture
		side.transparent_texture = NoTexture
		if exterior
			side.type = FULL_SIDE
			side.primary_texture = spanning_texture
		else
			side.transparent_texture = spanning_texture
			if upper_wall && lower_wall
				side.type = SPLIT_SIDE
				side.primary_texture = upper_texture
				side.secondary_texture = lower_texture
			elsif upper_wall
				side.type = HIGH_SIDE
				side.primary_texture = upper_texture
			elsif lower_wall
				side.type = LOW_SIDE
				side.primary_texture = lower_texture
			end
		end
		if is_control_panel
			side_flags |= 0x0002
		end
		side.flags = side_flags
		side.polygon = line.polygons[which_side]
		side.line = line
		line.sides[which_side] = side
		if panel_type >= 0
			side.permutation_type = Marathon::ControlPanelDefinitions[panel_type][1]
			perm_expr = attrs["control_panel_permutation"]
			side.permutation_name = ctx.eval_param(perm_expr)
			if side.permutation_type == 'terminal'
				term_file = side.permutation_name
				if term_file != ""
					add_terminal(term_file, line)
				end
			end
		end
		if @@DebugSides
			printf("Primary texture = %s\n", side.primary_texture.inspect)
			printf("Secondary texture = %s\n", side.secondary_texture.inspect)
			printf("Transparent texture = %s\n", side.transparent_texture.inspect)
		end
	end
	
	def texture_definition(wall_side, which_height)
		if wall_side
			material = wall_side.material
			#if material
				uv = wall_side.uvs[which_height]
				attrs = Marathon.get_wall_attributes(wall_side.face)
				ctx = wall_side.ctx
				mode = (attrs["transfer_mode"] or 0)
				light = ctx.eval_param(attrs["light"])
				return MarathonSideTextureDefinition.from_material(
					uv, material, mode, light)
			#end
		end
		return NoTexture
	end
	
	def add_terminal(name, line)
		if Marathon::DebugTerminal
			printf("Adding terminal %s\n", name.inspect)
		end
		if not @name_to_terminal.has_key?(name)
			n = @marathon_terminals.length
			term = Marathon::MarathonTerminal.new(n, name, line)
			@marathon_terminals.push(term)
			@name_to_terminal[name] = term
		end
	end

	#---------------------------------------------------------------------------------
	#
	#   Platform Creation
	#
	#   Create a Marathon Platform for each polygon of type "platform".
	#
	#---------------------------------------------------------------------------------

	def create_marathon_platforms()
		for poly in @marathon_polys
			if poly.type == 5 # _polygon_is_platform
				begin
					self.create_marathon_platform_for_poly(poly)
				rescue Marathon::ExportError
					error(poly.floor, poly.floor_ctx, $!.message)
				end
			end
		end
	end
	
	def create_marathon_platform_for_poly(poly)
		plat = new_marathon_platform(poly)
		self.adjust_adjacent_wall_texture_offsets(poly)
		ctx = poly.floor_ctx
		attrs = Marathon.get_platform_attributes(poly.floor)
		plat.flags = attrs["flags"] or 0
		plat.tag_name = ctx.eval_param(attrs["tag"])
	end
	
	def adjust_adjacent_wall_texture_offsets(poly)
		#  For platforms which have a moving ceiling, work out where the
		#  initial height of the ceiling will be and adjust the texture
		#  offsets of the upper wall faces of adjacent polygons to compensate.
		plat = poly.platform
		flags = plat.flags
		if flags & COMES_FROM_CEILING != 0
			adjacent = poly.adjacent_polys()
			floor_heights = []
			ceiling_heights = []
			for adj in adjacent
				if adj
					floor_heights.push(adj.floor_height)
					ceiling_heights.push(adj.ceiling_height)
				end
			end
			lowest_floor = floor_heights.min()
			#highest_floor = floor_heights.max()
			lowest_ceiling = ceiling_heights.min()
			highest_ceiling = ceiling_heights.max()
			if flags & COMES_FROM_FLOOR
				initial_ceiling = (lowest_floor + highest_ceiling) / 2
			else
				initial_ceiling = lowest_ceiling
			end
			offset = initial_ceiling - poly.ceiling_height
			for line in poly.lines
				if line
					side = line.side_for_other_poly(poly)
					if side
						side.adjust_upper_texture_offset(offset)
					end
				end
			end
		end
	end

	#---------------------------------------------------------------------------------
	#
	#   Object Creation
	#
	#   Create Marathon Objects from object markers and associate them with polys.
	#
	#---------------------------------------------------------------------------------

	def create_marathon_objects()
		for inst, ctx in @objects
			if @@DebugObjects
				printf("Creating Marathon object for %s instance %s\n",
					inst.definition.name, inst.name) ###
			end
			attrs = Marathon.get_object_attributes(inst)
			defn = inst.definition
			type_name, index_str = defn.description.split(",")
			type = MarathonObjectTypes[type_name]
			if not type
				error(inst, ctx,
					format('Unrecognised object type "%s".', type_name))
			end
			p, a = position_and_facing(inst, ctx)
			xy = marathon_coords([p.x, p.y])
			z = marathon_coord(p.z)
			poly = self.find_marathon_poly(xy, z)
			if poly == nil
				error(inst, ctx, "Unable to find containing polygon for object.")
			end
			if @@DebugObjects
				printf("In polygon %s\n", poly.number)
			end
			if type == 4 # goal
				index = goal_named(inst.name)
				if index < 0
					error(inst, ctx, "Name required for goal object.")
				end
			elsif type == 3 or type == 5 # player or sound
				index = (attrs["index"] or 0)
			else
				index = index_str.to_i()
			end
			if type == 0
				@monster_counts[index] += 1
			elsif type == 2
				@item_counts[index] += 1
			end
			flags = (attrs["flags"] or 0)
			activation = (attrs["activation"] or 0)
			obj = self.new_marathon_object()
			obj.type = type
			obj.index = index
			obj.facing = a
			obj.polygon = poly
			obj.location = [xy, z]
			obj.flags = flags | (activation << 14)
			obj.inst = inst
			obj.ctx = ctx
			poly.objects.push(obj)
			name = inst.name
			if name != ""
				if @name_to_marathon_object.has_key?(name)
					self.error(inst, ctx, format(
						"More than one object named '%s'.", name))
				end
				@name_to_marathon_object[name] = obj
			end
			if @@DebugObjects
				printf("Object list now %s\n", poly.objects.inspect)
			end
		end
	end
	
	def position_and_facing(inst, ctx)
		itrans = ctx.trans * inst.transformation
		p = itrans.origin
		v = itrans.xaxis
		if @@DebugObjects
			printf("Origin %s forward %s\n", p.inspect, v.inspect) ###
		end
		a = (Math.atan2(-v.y, v.x) * 0x100 / Math::PI).round
		if a < 0
			a += 0x200
		end
		return [p, a]
	end
	
	def find_marathon_poly(xy, z)
		#printf("Finding poly containing %s at height %s\n", xy.inspect, z.inspect) ###
		for poly in @marathon_polys
			if self.poly_contains_coords(poly, xy, z)
				return poly
			end
		end
		return nil
	end

	def poly_contains_coords(poly, xy, z)
		if z < poly.floor_height or z > poly.ceiling_height
			return false
		end
		x, y = xy
		for i in 0...poly.num_points
			j, k = poly.edge_point_numbers(i)
			(x0, y0) = @marathon_points[j]
			(x1, y1) = @marathon_points[k]
			d = (x - x0) * (y1 - y0) - (y - y0) * (x1 - x0)
			if d < 0
				return false
			end
		end
		return true
	end

	#---------------------------------------------------------------------------------
	#
	#   Terminal file parsing
	#
	#---------------------------------------------------------------------------------

	def parse_terminal_files
		for term in @marathon_terminals
			begin
				term.parse_file()
			rescue Errno::ENOENT
				printf("Warning: Terminal file '%s' not found.\n", term.filename)
			rescue Marathon::TerminalError
				line = term.representative_line
				wall = line.representative_wall
				error(wall.face, wall.ctx, $!.message)
			end
		end
	end

	#---------------------------------------------------------------------------------
	#
	#   Face Classification
	#
	#   Classify face as wall, floor or ceiling.
	#
	#---------------------------------------------------------------------------------
	
	def check_wall_face(face, ctx)
		if face.loops.length > 1
			error(face, ctx, "Wall face contains holes.")
		end
	end

	def check_floor_or_ceiling_face(face, ctx)
		#  TODO: This test turns out to be too fussy. Need to do it ourselves after
		#  mapping to marathon coords to avoid problems with floating point accuracy.
		#if not face.outer_loop.convex?
		#	error(face, "Floor or ceiling face is not convex.")
		#end
		if face.loops.length > 1
			error(face, ctx, "Floor or ceiling face contains holes.")
		end
		if face.vertices.length > 8
			error(face, ctx, "Floor or ceiling face has more than 8 corners.")
		end
	end
	
	def classify_face(face, ctx)
		normal = ctx.trans * face.normal
		#print("Face with normal", normal, "\n")
		z = normal.z
		if z.abs < 0.01
			#print("Wall found\n")
			check_wall_face(face, ctx)
			@walls.push([face, ctx])
		elsif z < -0.99
			#print("Floor found\n")
			check_floor_or_ceiling_face(face, ctx)
			@floors.push([face, ctx])
		elsif z > 0.99
			#print("Ceiling found\n")
			check_floor_or_ceiling_face(face, ctx)
			@ceilings.push([face, ctx])
		else
			error(face, ctx, "Face not classifiable as wall, floor or ceiling.")
		end
	end
	
	#---------------------------------------------------------------------------------
	#
	#   Model Processing
	#
	#   Process all the geometry of the model, expanding groups and component
	#   instances, extracting faces and object markers and classifying faces
	#   as walls, floors or ceilings.
	#
	#---------------------------------------------------------------------------------

	def add_light(name, attrs)
		n = @lights.length
		light = MarathonLight.new()
		@lights.push(light)
		@name_to_light_number[name] = n
		light.attrs = attrs
		return light
	end

	def process_light(inst, ctx)
		name = inst.name
		if name != ""
			if @name_to_light_number[name] != nil
				error(inst, ctx, format("More than one light named %s", name))
			end
			attrs = Marathon.get_light_attributes(inst)
			light = self.add_light(name, attrs)
			light.tag_name = ctx.eval_param(attrs["tag"])
		end
	end

	def add_media(name, inst, ctx)
		n = @media.length
		media = MarathonMedia.new()
		@media.push(media)
		@name_to_media_number[name] = n
		return media
	end

	def process_media(inst, ctx)
		name = inst.name
		if name != ""
			if @name_to_media_number[name] != nil
				error(inst, ctx, format("More than one liquid named %s", name))
			end
			attrs = Marathon.get_media_attributes(inst)
			media = self.add_media(name, inst, ctx)
			media.inst = inst
			media.ctx = ctx
			media.tide_name = ctx.eval_param(attrs["light_index"])
		end
	end

	def add_ambient_sound(name, inst, ctx)
		n = @ambient_sounds.length
		@ambient_sounds.push([inst, ctx])
		@name_to_ambient_sound_number[name] = n
	end

	def process_ambient_sound(inst, ctx)
		name = inst.name
		if name != ""
			if @name_to_ambient_sound_number[name] != nil
				error(inst, ctx, format("More than one ambient sound named %s", name))
			end
			self.add_ambient_sound(name, inst, ctx)
		end
	end

	def add_random_sound(name, inst, ctx)
		n = @random_sounds.length
		@random_sounds.push([inst, ctx])
		@name_to_random_sound_number[name] = n
	end

	def process_random_sound(inst, ctx)
		name = inst.name
		if name != ""
			if @name_to_random_sound_number[name] != nil
				error(inst, ctx, format("More than one random sound named %s", name))
			end
			self.add_random_sound(name, inst, ctx)
		end
	end
	
	def process_note(inst, ctx)
		@notes.push([inst, ctx])
	end

	def process_component_instance(inst, ctx)
		defn = inst.definition
		desc = defn.description
		if desc[0, 1] == "*"
			begin
				self.process_special_instance(desc, inst, ctx)
			rescue Marathon::ExportError
				error(inst, ctx, $!.message)
			end
		else
			ictx = Context.new(ctx, inst)
			self.process_entities(defn.entities, ictx)
		end
	end
	
	def process_special_instance(desc, inst, ctx)
		case desc
			when "*Light"
				if @@DebugLights
					printf("Classifying %s as a light\n", inst)
				end
				self.process_light(inst, ctx)
			when "*Liquid"
				if @@DebugMedia
					printf("Classifying %s as a liquid\n", inst)
				end
				self.process_media(inst, ctx)
			when "*AmbientSound"
				self.process_ambient_sound(inst, ctx)
			when "*RandomSound"
				self.process_random_sound(inst, ctx)
			when "*Note"
				self.process_note(inst, ctx)
			else
				@objects.push([inst, ctx])
		end
	end

	def process_group(grp, ctx)
		gctx = Context.new(ctx, grp)
		self.process_entities(grp.entities, gctx)
	end

	def process_entity(ent, ctx)
		#if @@Debug
		#	print("Processing ", ent, "\n")
		#end
		if ent.instance_of?(Sketchup::Face)
			self.classify_face(ent, ctx)
		elsif ent.instance_of?(Sketchup::ComponentInstance)
			old_context = @context
			@context = ent
			self.process_component_instance(ent, ctx)
			@context = old_context
		elsif ent.instance_of?(Sketchup::Group)
			self.process_group(ent, ctx)
		else
			#if @@Debug
			#	print("Skipping ", ent, "\n")
			#end
		end
	end

	def process_entities(entities, ctx)
		entities.each do |ent|
			if not @@VisibleOnly or ent.visible?
				self.process_entity(ent, ctx)
			end
		end
	end

	def process_space(entities)
		#trans = Geom::Transformation.new
		ctx = Context.new()
		self.process_entities(entities, ctx)
	end	

	#---------------------------------------------------------------------------------
	#
	#   Main Conversion Algorithm
	#
	#---------------------------------------------------------------------------------

	def export(model)
		self.delete_error_markers()
		self.process_space(model.entities)
		self.create_marathon_polys()
		self.create_marathon_lines_from_walls()
		self.create_marathon_lines_without_walls()
		self.create_marathon_sides()
		self.create_marathon_platforms()
		self.create_marathon_objects()
		self.parse_terminal_files()
		self.write_map_file(model.path)
	end

end # class MarathonExporter

#---------------------------------------------------------------------------------

def do_marathon_export()
	model = Sketchup.active_model
	exporter = MarathonExporter.new()
	catch :MarathonAbort do
		begin
			exporter.export(model)
		rescue Marathon::ExportError
			UI.beep
			UI.messagebox($!.message)
		end
	end
	exporter.dump()
	return nil
end

if (not $marathon_export_map)
	menu = marathon_submenu()
	menu.add_separator()
	menu.add_item("Export Map") {
		do_marathon_export()
	}
	$marathon_export_map = true
end
